import { Chance } from 'chance';
import { addDays, differenceInDays, isWeekend, startOfToday } from 'date-fns';

export interface Client {
    readonly id: string;
    readonly name: string;
}

export interface Role {
    readonly id: string;
    readonly name: string;
    readonly status: 'Active';
    readonly imagePath: string;
}

export interface People {
    readonly id: string;
    readonly firstName: string;
    readonly lastName: string;
    readonly roleId: string;
    readonly status: 'Active';
    readonly imagePath: string;
}

export interface Assignee {
    readonly id: string;
    readonly name: string;
    readonly sortName: string;
    readonly assigneeType: 'Person' | 'Role';
    readonly roleId: string;
    readonly status: 'Active';
    readonly imagePath: string;
}

export interface Task {
    readonly id: string;
    readonly name: string;
    readonly code: string;
}

export interface Project {
    readonly id: string;
    readonly name: string;
    readonly clientId: string;
    readonly code: string;
    readonly description: string;
    readonly status: 'Confirmed' | 'Tentative';
    readonly chargeable: boolean;
    readonly tasks: Task[];
}

export interface Assignment {
    readonly id: string;
    readonly associationId: string;
    readonly createdDate: string;
    readonly startDate: string;
    readonly endDate: string;
    readonly status: 'Confirmed' | 'Tentative';
    readonly hasNonWorkingDays: boolean;
    readonly contouredHours: { hours: number; date: string }[];
    readonly totalHours: number;
    readonly remainingHours: number;
    readonly utilizationPercentage: number;
}

export interface AssociationSummary {
    readonly projectId: string;
    readonly taskId: string;
    readonly assigneeId: string;
    readonly roleInstanceId: string;
    readonly associationId: string;
    readonly totalHours: number;
    readonly remainingHours: number;
    readonly startDate: string;
    readonly endDate: string;
    readonly createdDate: string;
}

export interface Tenant {
    readonly id: string;
    readonly name: string;
    readonly user: { userId: string; personId: string };
    readonly assignee: Assignee;
}

export class Data {
    private readonly chance = new Chance(1);

    public readonly tenant: Tenant;
    public readonly clients: Client[] = [];
    public readonly roles: Role[] = [];
    public readonly people: People[] = [];
    public readonly assignees: Assignee[] = [];
    public readonly projects: Project[] = [];
    public readonly assignments: Assignment[] = [];
    public readonly associationSummaries: AssociationSummary[] = [];

    constructor(opts: {
        clientCount: number;
        roleCount: number;
        peopleCount: number;
        projectCount: number;
        maxTasksPerProject: number;
        associationCount: number;
        maxAssignmentCountPerAssociation: number;
        assignmentStartDateOffset: { min: number; max: number };
        assignmentEndDateOffset: { min: number; max: number };
    }) {
        this.createClients(opts);
        this.createRoles(opts);
        this.createPeople(opts);
        this.createProjects(opts);
        this.createAssignments(opts);

        const person = this.chance.pickone(this.people);
        this.tenant = {
            id: this.chance.guid(),
            name: this.chance.company(),
            user: {
                userId: this.chance.guid(),
                personId: person.id,
            },
            assignee: {
                ...this.assignees.find((x) => x.id === person.id),
            },
        };
    }

    private createClients(opts: { clientCount: number }): void {
        for (let i = 0; i < opts.clientCount; i++) {
            this.clients.push({
                id: this.chance.guid(),
                name: this.chance.company(),
            });
        }
    }

    private createRoles(opts: { roleCount: number }): void {
        for (let i = 0; i < opts.roleCount; i++) {
            const role: Role = {
                id: this.chance.guid(),
                name: this.chance.animal(),
                status: 'Active',
                imagePath: null,
            };

            this.roles.push(role);

            this.assignees.push({
                id: role.id,
                name: role.name,
                sortName: role.name,
                assigneeType: 'Role',
                roleId: null,
                status: role.status,
                imagePath: role.imagePath,
            });
        }
    }

    private createPeople(opts: { peopleCount: number }): void {
        for (let i = 0; i < opts.peopleCount; i++) {
            const person: People = {
                id: this.chance.guid(),
                firstName: this.chance.first(),
                lastName: this.chance.last(),
                roleId: this.chance.pickone([
                    null,
                    ...this.roles.map((x) => x.id),
                ]),
                status: 'Active',
                imagePath: null,
            };

            this.people.push(person);

            this.assignees.push({
                id: person.id,
                name: `${person.firstName} ${person.lastName}`,
                sortName: `${person.lastName}${person.firstName}`,
                assigneeType: 'Person',
                roleId: null,
                status: person.status,
                imagePath: person.imagePath,
            });
        }
    }

    private createProjects(opts: {
        projectCount: number;
        maxTasksPerProject: number;
    }): void {
        for (let i = 0; i < opts.projectCount; i++) {
            const tasks: Task[] = [];

            const taskCount = this.chance.integer({
                min: 0,
                max: opts.maxTasksPerProject,
            });
            for (let j = 0; j < taskCount; j++) {
                tasks.push({
                    id: this.chance.guid(),
                    name: this.chance.animal(),
                    code: this.chance.word(),
                });
            }

            this.projects.push({
                id: this.chance.guid(),
                name: this.chance.animal(),
                clientId: this.chance.pickone([
                    null,
                    ...this.clients.map((x) => x.id),
                ]),
                code: this.chance.word(),
                description: this.chance.sentence(),
                status: this.chance.pickone<'Confirmed' | 'Tentative'>([
                    'Confirmed',
                    'Tentative',
                ]),
                chargeable: this.chance.bool(),
                tasks: tasks,
            });
        }
    }

    private createAssignments(opts: {
        associationCount: number;
        maxAssignmentCountPerAssociation: number;
        assignmentStartDateOffset: { min: number; max: number };
        assignmentEndDateOffset: { min: number; max: number };
    }): void {
        for (let i = 0; i < opts.associationCount; i++) {
            const project = this.chance.pickone(this.projects);
            const assignee = this.chance.pickone(this.assignees);
            const associationId = this.chance.guid();

            const startDate: Date = addDays(
                startOfToday(),
                this.chance.integer(opts.assignmentStartDateOffset)
            );
            const endDate: Date = addDays(
                startDate,
                this.chance.integer(opts.assignmentEndDateOffset)
            );

            const assignmentCount = this.chance.integer({
                min: 0,
                max: opts.maxAssignmentCountPerAssociation,
            });

            let totalAssociationHours = 0;
            let remainingAssociationHours = 0;

            for (let j = 0; j < assignmentCount; j++) {
                const contouredHours: { hours: number; date: string }[] = [];
                const hasNonWorkingDays: boolean = this.chance.bool();
                let totalHours = 0;
                let remainingHours = 0;

                const assignmentStartDate =
                    j == 0
                        ? startDate
                        : addDays(
                              startDate,
                              this.chance.integer({
                                  min: 0,
                                  max: differenceInDays(endDate, startDate),
                              })
                          );

                const assignmentEndDate =
                    j == assignmentCount
                        ? endDate
                        : addDays(
                              assignmentStartDate,
                              this.chance.integer({
                                  min: 0,
                                  max: differenceInDays(
                                      endDate,
                                      assignmentStartDate
                                  ),
                              })
                          );

                const hours = this.chance.floating({
                    min: 1,
                    max: 10,
                    fixed: 1,
                });
                for (
                    let date = assignmentStartDate;
                    date <= assignmentEndDate;
                    date = addDays(date, 1)
                ) {
                    if (hasNonWorkingDays || !isWeekend(date)) {
                        totalHours += hours;
                        totalAssociationHours += hours;
                        if (date >= startOfToday()) {
                            remainingHours += hours;
                            remainingAssociationHours += hours;
                        }

                        contouredHours.push({
                            hours: hours,
                            date: date.toISOString().split('T')[0],
                        });
                    }
                }

                this.assignments.push({
                    id: this.chance.guid(),
                    associationId: associationId,
                    createdDate: this.chance.date().toISOString(),
                    startDate: assignmentStartDate.toISOString().split('T')[0],
                    endDate: assignmentEndDate.toISOString().split('T')[0],
                    status: this.chance.pickone<'Confirmed' | 'Tentative'>([
                        'Confirmed',
                        'Tentative',
                    ]),
                    hasNonWorkingDays: hasNonWorkingDays,
                    contouredHours: contouredHours,
                    totalHours: totalHours,
                    remainingHours: remainingHours,
                    utilizationPercentage: totalHours / contouredHours.length,
                });
            }

            this.associationSummaries.push({
                projectId: project.id,
                taskId:
                    project.tasks.length == 0
                        ? null
                        : this.chance.pickone(project.tasks).id,
                assigneeId: this.chance.pickone(
                    this.assignees.map((x) => x.id)
                ),
                roleInstanceId:
                    assignee.assigneeType == 'Person'
                        ? '00000000-0000-0000-0000-000000000000'
                        : this.chance.guid(),
                associationId: associationId,
                totalHours: totalAssociationHours,
                remainingHours:
                    assignmentCount == 0 ? null : remainingAssociationHours,
                startDate:
                    assignmentCount == 0
                        ? null
                        : startDate.toISOString().split('T')[0],
                endDate:
                    assignmentCount == 0
                        ? null
                        : endDate.toISOString().split('T')[0],
                createdDate: this.chance.date().toISOString(),
            });
        }
    }
}
