import * as path from 'path';
import * as fs from 'fs';

import { codeFrameColumns } from '@babel/code-frame';
import { TestCase, TestResult, TestStep, Location } from './testReporter';

import { FullConfig, TestError } from '@playwright/test';

// eslint-disable-next-line @typescript-eslint/no-var-requires
const StackUtils = require('stack-utils');

const stackUtils = new StackUtils();

export type TestResultOutput = {
    chunk: string | Buffer;
    type: 'stdout' | 'stderr';
};
export const kOutputSymbol = Symbol('output');

type ErrorDetails = {
    message: string;
    location?: Location;
};

export const formatFailure = (
    config: FullConfig,
    test: TestCase,
    options: {
        index?: number;
        includeStdio?: boolean;
        includeAttachments?: boolean;
    } = {}
): {
    message: string;
} => {
    const { index, includeStdio, includeAttachments = true } = options;
    const lines: string[] = [];
    const header = formatTestHeader(config, test, '  ', index);
    lines.push(header);
    for (const result of test.results) {
        const resultLines: string[] = [];
        const errors = formatResultFailure(config, test, result, '    ');
        if (!errors.length) continue;
        const retryLines = [];
        if (result.retry) {
            retryLines.push('');
            retryLines.push(pad(`    Retry #${result.retry}`, '-'));
        }
        resultLines.push(...retryLines);
        resultLines.push(...errors.map((error) => '\n' + error.message));
        if (includeAttachments) {
            for (let i = 0; i < result.attachments.length; ++i) {
                const attachment = result.attachments[i];
                const hasPrintableContent =
                    attachment.contentType.startsWith('text/') &&
                    attachment.body;
                if (!attachment.path && !hasPrintableContent) continue;
                resultLines.push('');
                resultLines.push(
                    pad(
                        `    attachment #${i + 1}: ${attachment.name} (${
                            attachment.contentType
                        })`,
                        '-'
                    )
                );
                if (attachment.path) {
                    const relativePath = path.relative(
                        process.cwd(),
                        attachment.path
                    );
                    resultLines.push(`    ${relativePath}`);
                    // Make this extensible
                    if (attachment.name === 'trace') {
                        resultLines.push(`    Usage:`);
                        resultLines.push('');
                        resultLines.push(
                            `        npx playwright show-trace ${relativePath}`
                        );
                        resultLines.push('');
                    }
                } else {
                    if (
                        attachment.contentType.startsWith('text/') &&
                        attachment.body
                    ) {
                        let text = attachment.body.toString();
                        if (text.length > 300)
                            text = text.slice(0, 300) + '...';
                        resultLines.push(`    ${text}`);
                    }
                }
                resultLines.push(pad('   ', '-'));
            }
        }
        const output = ((result as any)[kOutputSymbol] ||
            []) as TestResultOutput[];
        if (includeStdio && output.length) {
            const outputText = output
                .map(({ chunk, type }) => {
                    const text = chunk.toString('utf8');
                    if (type === 'stderr') return stripAnsiEscapes(text);
                    return text;
                })
                .join('');
            resultLines.push('');
            resultLines.push(
                pad('--- Test output', '-') +
                    '\n\n' +
                    outputText +
                    '\n' +
                    pad('', '-')
            );
        }

        lines.push(...resultLines);
    }
    lines.push('');
    return {
        message: lines.join('\n'),
    };
};

const formatResultFailure = (
    config: FullConfig,
    test: TestCase,
    result: TestResult,
    initialIndent: string
): ErrorDetails[] => {
    const errorDetails: ErrorDetails[] = [];

    if (result.status === 'passed' && test.expectedStatus === 'failed') {
        errorDetails.push({
            message: indent(`Expected to fail, but passed.`, initialIndent),
        });
    }

    for (const error of result.errors) {
        const formattedError = formatError(config, error, test.location.file);
        errorDetails.push({
            message: indent(formattedError.message, initialIndent),
            location: formattedError.location,
        });
    }
    return errorDetails;
};

const relativeFilePath = (config: FullConfig, file: string): string => {
    return path.relative(config.rootDir, file) || path.basename(file);
};

const relativeTestPath = (config: FullConfig, test: TestCase): string => {
    return relativeFilePath(config, test.location.file);
};

const stepSuffix = (step: TestStep | undefined) => {
    const stepTitles = step ? step.titlePath() : [];
    return stepTitles.map((t) => ' › ' + t).join('');
};

const formatTestTitle = (
    config: FullConfig,
    test: TestCase,
    step?: TestStep
): string => {
    // root, project, file, ...describes, test
    const [, projectName, , ...titles] = test.titlePath();
    const location = `${relativeTestPath(config, test)}:${test.location.line}:${
        test.location.column
    }`;
    const projectTitle = projectName ? `[${projectName}] › ` : '';
    return `${projectTitle}${location} › ${titles.join(' › ')}${stepSuffix(
        step
    )}`;
};

const formatTestHeader = (
    config: FullConfig,
    test: TestCase,
    indent: string,
    index?: number
): string => {
    const title = formatTestTitle(config, test);
    const header = `${indent}${index ? index + ') ' : ''}${title}`;
    return pad(header, '=');
};

const formatError = (
    config: FullConfig,
    error: TestError,
    file?: string
): ErrorDetails => {
    const stack = error.stack;
    const tokens = [];
    let location: Location | undefined;
    if (stack) {
        const parsed = prepareErrorStack(stack, file);
        tokens.push(parsed.message);
        location = parsed.location;
        if (location) {
            try {
                const source = fs.readFileSync(location.file, 'utf8');
                const codeFrame = codeFrameColumns(
                    source,
                    { start: location },
                    { highlightCode: false }
                );
                // Convert /var/folders to /private/var/folders on Mac.
                if (!file || fs.realpathSync(file) !== location.file) {
                    tokens.push('');
                    tokens.push(
                        `   at ` +
                            `${relativeFilePath(config, location.file)}:${
                                location.line
                            }`
                    );
                }
                tokens.push('');
                tokens.push(codeFrame);
            } catch (e) {
                // Failed to read the source file - that's ok.
            }
        }
        tokens.push('');
        tokens.push(parsed.stackLines.join('\n'));
    } else if (error.message) {
        tokens.push(error.message);
    } else if (error.value) {
        tokens.push(error.value);
    }
    return {
        location,
        message: tokens.join('\n'),
    };
};

const pad = (line: string, char: string): string => {
    if (line) line += ' ';
    return line + char.repeat(Math.max(0, 100 - line.length));
};

const indent = (lines: string, tab: string) => {
    return lines.replace(/^(?=.+$)/gm, tab);
};

const prepareErrorStack = (
    stack: string,
    file?: string
): {
    message: string;
    stackLines: string[];
    location?: Location;
} => {
    if (file) {
        // Stack will have /private/var/folders instead of /var/folders on Mac.
        file = fs.realpathSync(file);
    }
    const lines = stack.split('\n');
    let firstStackLine = lines.findIndex((line) => line.startsWith('    at '));
    if (firstStackLine === -1) firstStackLine = lines.length;
    const message = lines.slice(0, firstStackLine).join('\n');
    const stackLines = lines.slice(firstStackLine);
    let location: Location | undefined;
    for (const line of stackLines) {
        const parsed = stackUtils.parseLine(line);
        if (!parsed || !parsed.file) continue;
        const resolvedFile = path.join(process.cwd(), parsed.file);
        if (!file || resolvedFile === file) {
            location = {
                file: resolvedFile,
                column: parsed.column || 0,
                line: parsed.line || 0,
            };
            break;
        }
    }
    return { message, stackLines, location };
};

const ansiRegex = new RegExp(
    /* eslint-disable max-len */
    /* eslint-disable no-control-regex */
    '[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:[a-zA-Z\\d]*(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]*)*)?\\u0007)|(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PR-TZcf-ntqry=><~]))',
    /* eslint-enable no-control-regex */
    /* eslint-enable max-len */
    'g'
);
export const stripAnsiEscapes = (str: string): string => {
    return str.replace(ansiRegex, '');
};
