import * as fs from 'fs';
import * as path from 'path';
import {
    FullConfig,
    Reporter,
    Suite,
    TestCase,
} from '@playwright/test/reporter';
import { formatFailure, stripAnsiEscapes } from './base';

class NUnit3Reporter implements Reporter {
    private config!: FullConfig;
    private suite!: Suite;
    private startTime: Date;
    private outputFile: string | undefined;
    private stripANSIControlSequences = false;

    constructor(
        options: {
            outputFile?: string;
            stripANSIControlSequences?: boolean;
        } = {}
    ) {
        this.outputFile = options.outputFile;
        this.stripANSIControlSequences =
            options.stripANSIControlSequences || false;
    }

    printsToStdio() {
        return !this.outputFile;
    }

    onBegin(config: FullConfig, suite: Suite) {
        this.config = config;
        this.suite = suite;
        this.startTime = new Date();
    }

    onEnd() {
        const endTime = new Date();
        const children: XMLEntry[] = [];
        for (const projectSuite of this.suite.suites) {
            for (const fileSuite of projectSuite.suites)
                children.push(this._buildTestSuite(fileSuite));
        }
        const tokens: string[] = [];

        const root: XMLEntry = {
            name: 'test-run',
            attributes: {
                'start-time': this.startTime.toISOString(),
                'end-time': endTime.toISOString(),
            },
            children,
        };

        serializeXML(root, tokens, this.stripANSIControlSequences);
        const reportString = tokens.join('\n');
        if (this.outputFile) {
            fs.mkdirSync(path.dirname(this.outputFile), { recursive: true });
            fs.writeFileSync(this.outputFile, reportString);
        } else {
            console.log(reportString);
        }
    }

    private _buildTestSuite(suite: Suite): XMLEntry {
        const children: XMLEntry[] = [];

        suite.allTests().forEach((test: TestCase) => {
            this._addTestCase(test, children);
        });

        const entry: XMLEntry = {
            name: 'test-suite',
            attributes: {
                name: suite.location
                    ? path.relative(this.config.rootDir, suite.location.file)
                    : '',
                type: 'Assembly',
            },
            children,
        };

        return entry;
    }

    private _addTestCase(test: TestCase, entries: XMLEntry[]) {
        let startDate: Date;
        for (const result of test.results) {
            if (!startDate || startDate > result.startTime) {
                startDate = result.startTime;
            }
        }

        if (!startDate) {
            startDate = this.startTime;
        }

        const duration = test.results.reduce(
            (acc, value) => acc + value.duration,
            0
        );

        const endDate = new Date(startDate.getTime() + duration);

        let result: string;

        if (test.outcome() === 'skipped') {
            result = 'Skipped';
        } else if (!test.ok()) {
            result = 'Failed';
        } else {
            result = 'Passed';
        }

        const attachments: XMLEntry = {
            name: 'attachments',
            children: [] as XMLEntry[],
        };

        for (const result of test.results) {
            for (const attachment of result.attachments) {
                if (!attachment.path) continue;

                attachments.children.push({
                    name: 'attachment',
                    children: [
                        {
                            name: 'filePath',
                            text: attachment.path,
                        },
                    ],
                });
            }
        }

        const entry = {
            name: 'test-case',
            attributes: {
                fullname: test.titlePath().slice(3).join(' '),
                // Skip root, project, file
                name: test.titlePath().slice(3).join(' '),
                'start-time': startDate.toISOString(),
                'end-time': endDate.toISOString(),
                result: result,
            },
            children: [] as XMLEntry[],
        };

        if (attachments.children.length) {
            entry.children.push(attachments);
        }

        entries.push(entry);

        if (!test.ok()) {
            const systemOut: string[] = [];
            const systemErr: string[] = [];
            for (const result of test.results) {
                systemOut.push(...result.stdout.map((item) => item.toString()));
                systemErr.push(...result.stderr.map((item) => item.toString()));
            }

            let output = '';
            if (systemOut.length) output += `std-out: ${systemOut.join('')}`;

            if (systemErr.length) {
                if (output.length) {
                    output += '\n';
                }
                output += `std-err: ${systemErr.join('')}`;
            }

            const failure: XMLEntry = {
                name: 'failure',
                children: [
                    {
                        name: 'message',
                        text: `${path.basename(test.location.file)}:${
                            test.location.line
                        }:${test.location.column} ${test.title}`,
                    },
                    {
                        name: 'stack-trace',
                        text: stripAnsiEscapes(
                            formatFailure(this.config, test).message
                        ),
                    },
                ],
            };

            if (output.length) {
                failure.children.push({
                    name: 'output',
                    text: output,
                });
            }

            entry.children.push(failure);
        }
    }
}

type XMLEntry = {
    name: string;
    attributes?: { [name: string]: string | number | boolean };
    children?: XMLEntry[];
    text?: string;
};

const serializeXML = (
    entry: XMLEntry,
    tokens: string[],
    stripANSIControlSequences: boolean
) => {
    const attrs: string[] = [];
    for (const [name, value] of Object.entries(entry.attributes || {}))
        attrs.push(
            `${name}="${escape(String(value), stripANSIControlSequences)}"`
        );
    tokens.push(`<${entry.name}${attrs.length ? ' ' : ''}${attrs.join(' ')}>`);

    for (const child of entry.children || []) {
        serializeXML(child, tokens, stripANSIControlSequences);
    }

    if (entry.text) {
        let text = entry.text;
        if (stripANSIControlSequences) text = stripAnsiEscapes(text);
        tokens.push(`<![CDATA[${text.replace(']]>', ']]]]><!CDATA[>')}]]>`);
    }
    tokens.push(`</${entry.name}>`);
};

// See https://en.wikipedia.org/wiki/Valid_characters_in_XML
const discouragedXMLCharacters =
    // eslint-disable-next-line no-control-regex
    /[\u0001-\u0008\u000b-\u000c\u000e-\u001f\u007f-\u0084\u0086-\u009f]/g;

const escape = (text: string, stripANSIControlSequences: boolean): string => {
    if (stripANSIControlSequences) text = stripAnsiEscapes(text);
    const escapeRe = /[&"<>]/g;
    text = text.replace(
        escapeRe,
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        (c) => ({ '&': '&amp;', '"': '&quot;', '<': '&lt;', '>': '&gt;' }[c]!)
    );

    text = text.replace(discouragedXMLCharacters, '');
    return text;
};

export default NUnit3Reporter;
