/**
 * Scheduler 1 script file
 */
import { Component, OnDestroy, AfterViewInit, ViewChild } from '@angular/core';
import { Store } from '@ngrx/store';
import {
    ActivatedRoute,
    Event,
    NavigationEnd,
    NavigationStart,
    Router,
} from '@angular/router';
import { BryntumSchedulerProComponent } from '@bryntum/schedulerpro-angular';
import {
    DateHelper,
    SchedulerPro,
    SchedulerProConfig,
    SchedulerEventModel,
    SchedulerResourceModel,
    ViewPreset,
} from '@bryntum/schedulerpro';
import { scheduler1Config } from './scheduler1.config';
import events from '../../assets/data/scheduler1/ProjectPhase-12500-events.json';
import resources from '../../assets/data/scheduler1/ProjectPhase-12500-resources.json';

@Component({
    selector: 'app-scheduler1',
    templateUrl: './scheduler1.component.html',
    styleUrls: ['./scheduler1.component.scss'],
})
export class Scheduler1Component implements AfterViewInit, OnDestroy {
    @ViewChild('schedulerpro')
    schedulerComponent: BryntumSchedulerProComponent;
    public schedulerConfig: Partial<SchedulerProConfig> = scheduler1Config;
    private scheduler: SchedulerPro;
    public timelineView: string = 'DAY';
    public resources$: Partial<SchedulerResourceModel>[];
    public events$: Partial<SchedulerEventModel>[];
    public expandCollapseButtonText: string = 'EXPAND ALL';

    private expanded: boolean = false;

    // This is what's wired up to the UI feature toggles
    // and is meant to be a simple solution non-complex
    // implementation.
    public featureToggles: {
        allowEventDrag: boolean;
        allowEventResize: boolean;
        allowEventDragCreate: boolean;
        allowRowReorder: boolean;
    } = {
        allowEventDrag: true,
        allowEventResize: true,
        allowEventDragCreate: true,
        allowRowReorder: true,
    };

    constructor(
        public router: Router,
        public route: ActivatedRoute,
        public store: Store<{
            barMargin: { barMargin: number };
        }>
    ) {
        performance.mark('setResources::start');
        this.resources$ = resources;
        performance.mark('setResources::end');
        performance.measure(
            'setResources',
            'setResources::start',
            'setResources::end'
        );

        performance.mark('setEvents::start');
        this.events$ = events;
        performance.mark('setEvents::end');
        performance.measure('setEvents', 'setEvents::start', 'setEvents::end');

        console.log('resources count:', resources.length);
        console.log('event count:', events.length);
    }

    public setTimelineView(viewMode: string): void {
        performance.mark('setTimelineView::start');
        this.timelineView = viewMode;
        this.scheduler.viewPreset = this.getViewPreset(this.timelineView);
        performance.mark('setTimelineView::end');
        performance.measure(
            'setTimelineView',
            'setTimelineView::start',
            'setTimelineView::end'
        );
    }

    public expandCollapseAll(): void {
        if (!this.expanded) {
            this.expandCollapseButtonText = 'COLLAPSE ALL';
            performance.mark('expandCollapseAll::expandAll::start');
            this.scheduler.expandAll();
            performance.mark('expandCollapseAll::expandAll::end');
            performance.measure(
                'expandCollapseAll::expandAll',
                'expandCollapseAll::expandAll::start',
                'expandCollapseAll::expandAll::end'
            );
            this.expanded = true;
        } else {
            performance.mark('expandCollapseAll::collapseAll::start');
            this.scheduler.collapseAll();
            performance.mark('expandCollapseAll::collapseAll::end');
            performance.measure(
                'expandCollapseAll::collapseAll',
                'expandCollapseAll::collapseAll::start',
                'expandCollapseAll::collapseAll::end'
            );
            this.expanded = false;
            this.expandCollapseButtonText = 'EXPAND ALL';
        }
    }

    public goToToday(): void {
        performance.mark('goToToday::start');
        this.scheduler
            .scrollToNow({
                block: 'center',
            })
            .then(() => {
                performance.mark('goToToday::end');
                performance.measure(
                    'goToToday',
                    'goToToday::start',
                    'goToToday::start'
                );
            });
    }

    /************************************************************
     ** Here are the handlers for the UI feature toggles.
     * ***********************************************************/
    public onAllowEventDragToggle(event): void {
        this.scheduler.features['eventDrag'].disabled = !event.checked;
        this.featureToggles.allowEventDrag = event.checked;
    }

    public onAllowEventResizeToggle(event): void {
        this.scheduler.features['eventResize'].disabled = !event.checked;
        this.featureToggles.allowEventResize = event.checked;
    }

    public onAllowEventDragCreateToggle(event): void {
        this.scheduler.features['eventDragCreate'].disabled = !event.checked;
        this.featureToggles.allowEventDragCreate = event.checked;
    }

    public onAllowRowReorderToggle(event): void {
        this.scheduler.features['rowReorder'].disabled = !event.checked;
        this.featureToggles.allowRowReorder = event.checked;
    }

    ngAfterViewInit(): void {
        this.scheduler = this.schedulerComponent.instance;

        performance.mark('ngAfterViewInit::viewPreset::start');
        this.scheduler.viewPreset = this.getViewPreset(this.timelineView);
        performance.mark('ngAfterViewInit::viewPreset::end');
        performance.measure('ngAfterViewInit::viewPreset');

        const saveState = (event: Event) => {
            if (event instanceof NavigationStart) {
                if (this.router.isActive(this.route.routeConfig.path, true)) {
                    this.scheduler.storeScroll();
                }
            } else if (event instanceof NavigationEnd) {
                if (this.router.isActive(this.route.routeConfig.path, true)) {
                    this.scheduler.restoreScroll();
                }
            }
        };

        this.router.events.subscribe(saveState);
        this.goToToday();
    }

    ngOnDestroy(): void {}

    private getViewPreset(mode: string): ViewPreset {
        switch (mode) {
            case 'DAY':
                return new ViewPreset({
                    id: 'dayViewPreset',
                    tickWidth: 65,
                    shiftIncrement: 1,
                    shiftUnit: 'day',
                    timeResolution: {
                        unit: 'day',
                        increment: 1,
                    },
                    headers: [
                        {
                            unit: 'month',
                            renderer: (start): string => {
                                return DateHelper.format(start, 'MMM YYYY');
                            },
                            align: 'end',
                        },
                        {
                            unit: 'day',
                            dateFormat: 'DD',
                        },
                    ],

                    columnLinesFor: 0,
                });
            case 'WEEK':
                return new ViewPreset({
                    id: 'weekViewPreset',
                    tickWidth: 75,
                    shiftIncrement: 1,
                    shiftUnit: 'day',
                    timeResolution: {
                        unit: 'week',
                        increment: 1,
                    },
                    headers: [
                        {
                            unit: 'week',
                            renderer: (start): string => {
                                const firstDayOfMonth =
                                    DateHelper.getFirstDateOfMonth(start);
                                const weekNumberOfFirstDay =
                                    DateHelper.getWeekNumber(
                                        firstDayOfMonth
                                    )[1];
                                const currentDayWeekNumber =
                                    DateHelper.getWeekNumber(start)[1];

                                if (
                                    weekNumberOfFirstDay ===
                                    currentDayWeekNumber
                                ) {
                                    return DateHelper.format(start, 'MMM YYYY');
                                } else {
                                    return '';
                                }
                            },
                        },
                        {
                            unit: 'week',
                            headerCellCls: 'week-view-preset-header-1',
                            increment: 1,
                            renderer: (start, end): string => {
                                const startPeriod = DateHelper.format(
                                    start,
                                    'DD'
                                );
                                const endPeriod = DateHelper.format(
                                    DateHelper.add(end, -1, 'day'),
                                    'DD'
                                );
                                return `${startPeriod}-${endPeriod}`;
                            },
                        },
                    ],

                    columnLinesFor: 0,
                });

            case 'MONTH':
                return new ViewPreset({
                    id: 'monthViewPreset',
                    tickWidth: 76,
                    shiftIncrement: 1,
                    shiftUnit: 'month',
                    timeResolution: {
                        unit: 'month',
                        increment: 1,
                    },
                    headers: [
                        {
                            unit: 'year',
                            renderer: (start): string => {
                                if (DateHelper.get(start, 'month') === 0) {
                                    return DateHelper.format(start, 'YYYY');
                                } else {
                                    return '';
                                }
                            },
                        },
                        {
                            unit: 'month',
                            renderer: (startDate): string => {
                                return DateHelper.format(startDate, 'MMM');
                            },
                        },
                    ],

                    columnLinesFor: 0,
                });
        }
    }
}
