import { Chance } from 'Chance';
import { addDays, startOfToday } from 'date-fns';
import * as fs from 'fs';

enum ResourceType {
    Phase = 'Phase',
    Project = 'Project',
    Assignee = 'Assignee',
}

export enum ScheduleViewType {
    ProjectPhase = 'ProjectPhase',
}

export class DataGeneratorService {
    private readonly chance = new Chance();
    private readonly resources = [];
    private readonly events = [];

    private readonly opts = {
        assignmentStartDateOffset: { min: -180, max: 365 },
        assignmentEndDateOffset: { min: 1, max: 180 },
        maxParentResourceCount: 50,
        maxChildResourceCount: 5,
        maxResourceEvents: 10,
    };

    constructor() {}

    private generateResources(scheduleType) {
        let currentCount = 0;

        while (currentCount < this.opts.maxParentResourceCount) {
            this.resources.push(
                this.generateResource(scheduleType, ResourceType.Project, null)
            );
            currentCount++;
        }
    }

    private generateEvents(resourceId: string) {
        let currentCount = 0;
        const maxEventCount = this.opts.maxResourceEvents;

        while (currentCount < maxEventCount) {
            const startDate: Date = addDays(
                startOfToday(),
                this.chance.integer(this.opts.assignmentStartDateOffset)
            );
            const endDate: Date = addDays(
                startDate,
                this.chance.integer(this.opts.assignmentEndDateOffset)
            );
            const newEvent = {
                id: this.chance.guid(),
                name: this.chance.name(),
                editable: true,
                draggable: true,
                startDate,
                endDate,
                eventType: 'HourEntry',
                resourceId,
            };

            this.events.push(newEvent);
            currentCount++;
        }
    }

    private generateResource(
        scheduleViewType: ScheduleViewType,
        resourceType: ResourceType,
        parentId: string | null
    ) {
        const id = this.chance.guid();
        const resourceName =
            resourceType + ' ' + this.chance.word({ length: 6 });
        const childCount = this.opts.maxChildResourceCount;

        let newResource = {
            resourceType,
            id,
            name: resourceName,
        };

        if (parentId) {
            newResource['resourceId'] = parentId;
        }

        switch (scheduleViewType) {
            case ScheduleViewType.ProjectPhase:
                if (
                    resourceType === ResourceType.Project ||
                    resourceType === ResourceType.Phase
                ) {
                    let currentChildCount = 0;

                    newResource['children'] = [];
                    while (currentChildCount < childCount) {
                        newResource['children'].push(
                            this.generateResource(
                                scheduleViewType,
                                resourceType === ResourceType.Project
                                    ? ResourceType.Phase
                                    : ResourceType.Assignee,
                                id
                            )
                        );
                        currentChildCount++;
                    }
                } else if (resourceType === ResourceType.Assignee) {
                    this.generateEvents(id);
                }
                break;
        }

        return newResource;
    }

    public generate(scheduleType: ScheduleViewType) {
        this.generateResources(scheduleType);

        try {
            const path = require('path');
            const projectRoot = path.resolve(__dirname, '..');
            const resourcesOutputPath = path.join(
                projectRoot,
                `src/assets/data/scheduler1/${scheduleType}-${this.events.length}-resources.json`
            );
            const eventsOutputPath = resourcesOutputPath.replace(
                'resources',
                'events'
            );

            fs.writeFileSync(
                resourcesOutputPath,
                JSON.stringify(this.resources)
            );
            fs.writeFileSync(eventsOutputPath, JSON.stringify(this.events));

            console.log('events total:', this.events.length);
        } catch (err) {
            console.error(err);
        }
    }
}
